/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dao.filesystem;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;

import org.apache.log4j.Logger;

import br.registro.dnsshim.common.repository.FileSystemUtil;
import br.registro.dnsshim.xfrd.domain.Slave;
import br.registro.dnsshim.xfrd.domain.SlaveGroup;
import br.registro.dnsshim.xfrd.domain.XfrdConfig;
import br.registro.dnsshim.xfrd.domain.Zonename;
import br.registro.dnsshim.xfrd.domain.logic.XfrdConfigManager;
import br.registro.dnsshim.xfrd.repository.SlaveGroupRepository;

public class SlaveGroupDao implements SlaveGroupRepository {
	private static final Logger logger = Logger.getLogger(SlaveGroupDao.class);

	private static final String EXTENSION = ".slaves";
	private static final String REMOVED_EXTENSION = ".removed";
	
	public static final String REMOVED_COMMAND_EXTENSION = REMOVED_EXTENSION + ".rndc.";
	private static final String REMOVED_ERROR_EXTENSION = REMOVED_COMMAND_EXTENSION + "err";
	
	private static final String ADDED_EXTENSION = ".added";
	
	public static final String ADDED_COMMAND_EXTENSION = ADDED_EXTENSION + ".rndc.";
	private static final String ADDED_ERROR_EXTENSION = ADDED_COMMAND_EXTENSION + "err";
	private final String BASE_DIR;
	
	private static final String ZONES_PREFIX = "zones="; 

	public SlaveGroupDao() {
		StringBuilder root = new StringBuilder();
		try {
			// {DNSSHIM_HOME|user.home}/dnsshim/conf/xfrd/slaves/<groupname>.slaves
			if (System.getenv("DNSSHIM_HOME") == null) {
				root.append(System.getProperty("user.home"));
				root.append(File.separatorChar);
				root.append("dnsshim");
			} else {
				root.append(System.getenv("DNSSHIM_HOME"));
			}
			root.append(File.separatorChar);
			root.append("xfrd");
			root.append(File.separatorChar);
			root.append("conf");
			root.append(File.separatorChar);
			root.append("slaves");
			root.append(File.separatorChar);

			File directory = new File(root.toString());
			FileSystemUtil.makePath(directory);
			
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		BASE_DIR = root.toString();
	}
	
	@Override
	public void save(SlaveGroup slaveGroup) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving SlaveGroup...");
		}
		
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(EXTENSION);
		
		File file = new File(path.toString());
		Map<String, Slave> slaves = slaveGroup.getSlaves();
		
		StringBuilder content = new StringBuilder();

		content.append(ZONES_PREFIX);
		for (Zonename zonename : slaveGroup.getZonenames()) {
			content.append(zonename.getName());
			content.append(';');
		}
		content.append(System.getProperty("line.separator"));

		for (Entry<String, Slave> entry : slaves.entrySet()) {
			Slave slave = entry.getValue();
			content.append(slave.getAddress().getHostAddress());
			content.append(' ');
			content.append(slave.getPort());
			content.append(System.getProperty("line.separator"));			
		}
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(file));
		bw.write(content.toString());
		bw.close();
	}

	@Override
	public SlaveGroup findByName(String name) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(name);
		path.append(EXTENSION);
		
		File file = new File(path.toString());
		if (!file.exists()) {
			return null;
		}
	
		return parse(file);
	}

	@Override
	public boolean delete(String name) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(name);
		path.append(EXTENSION);
		
		File file = new File(path.toString());
		return file.delete();
	}

	@Override
	public List<SlaveGroup> findAll() throws IOException {
		List<SlaveGroup> groups = new ArrayList<SlaveGroup>();
		SlaveGroup slaveGroup = null;
		
		File groupPath = new File(BASE_DIR);
		File[] list = groupPath.listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return (name.endsWith(EXTENSION));
			}
		});
		
		for (File file : list) {
			slaveGroup = parse(file);
			groups.add(slaveGroup);
		}
		
		groupPath = new File(BASE_DIR);
		list = groupPath.listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return (name.endsWith(REMOVED_EXTENSION));
			}
		});
		
		if (list.length > 0 && slaveGroup == null) {
			throw new IOException("Invalid configuration. Removed zones file without slave group file.");
		}
		
		for (File file : list) {
			File tmp = new File(file.getAbsolutePath()+".tmp");
			file.renameTo(tmp);
			parseRemovedZoneFile(tmp, slaveGroup);
			file.delete();
			tmp.delete();
		}
		
		list = groupPath.listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return (name.endsWith(ADDED_EXTENSION));
			}
		});
		
		if (list.length > 0 && slaveGroup == null) {
			throw new IOException("Invalid configuration. Added zones file without slave group file.");
		}
		
		for (File file : list) {
			File tmp = new File(file.getAbsolutePath()+".tmp");
			file.renameTo(tmp);
			parseAddedZoneFile(tmp, slaveGroup);
			file.delete();
			tmp.delete();
		}
		
		
		
		return groups;
	}
	
	private SlaveGroup parseAddedZoneFile(File file, SlaveGroup slaveGroup) throws FileNotFoundException, UnknownHostException {
		Scanner scanner = new Scanner(file);
		scanner.useDelimiter(System.getProperty("line.separator"));
		while (scanner.hasNext()) {
			String line = scanner.next();
			Scanner lineScanner = new Scanner(line);
			String zone = lineScanner.next(); // first word in line
			if (!zone.isEmpty()) {
				Zonename zonename = new Zonename(zone.trim());
				slaveGroup.addZonename(zonename, true);
			}
		}
		return slaveGroup;
	}
	
	private SlaveGroup parseRemovedZoneFile(File file, SlaveGroup slaveGroup) throws FileNotFoundException, UnknownHostException {
		Scanner scanner = new Scanner(file);
		scanner.useDelimiter(System.getProperty("line.separator"));
		while (scanner.hasNext()) {
			String line = scanner.next();
			Scanner lineScanner = new Scanner(line);
			String zone = lineScanner.next(); // first word in line
			if (!zone.isEmpty()) {
				Zonename zonename = new Zonename(zone.trim());
				slaveGroup.removeZonename(zonename);
			}
		}
		return slaveGroup;
	}

	private SlaveGroup parse(File file) throws FileNotFoundException, UnknownHostException {
		Scanner scanner = new Scanner(file);
		scanner.useDelimiter(System.getProperty("line.separator"));
		String groupName = file.getName().substring(0, file.getName().length() - EXTENSION.length());
		SlaveGroup slaveGroup = new SlaveGroup(); 
		slaveGroup.setName(groupName);
		
		String line = scanner.next();
		Scanner lineScanner = null;
		if (line.startsWith(ZONES_PREFIX)) {
			line = line.substring(ZONES_PREFIX.length());
			lineScanner = new Scanner(line).useDelimiter(";");
			while (lineScanner.hasNext()) {
				String zone = lineScanner.next();
				Zonename zonename = new Zonename(zone);
				slaveGroup.addZonename(zonename, false);
			}
		}
		
		while (scanner.hasNext()) {
			line = scanner.next();
			lineScanner = new Scanner(line).useDelimiter(" ");
			String address = lineScanner.next();
			int port = Integer.parseInt(lineScanner.next());
			
			Slave slave = new Slave();
			slave.setAddress(InetAddress.getByName(address));
			slave.setPort(port);
			slaveGroup.addSlave(slave);
		}
		return slaveGroup;
	}

	@Override
	public void saveAddedZones(SlaveGroup slaveGroup) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving added zones...");
		}
		
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(ADDED_EXTENSION);
		
		File file = new File(path.toString());
		StringBuilder content = new StringBuilder();

		for (Zonename zonename : slaveGroup.getAddedZonenames()) {
			content.append(zonename.getName());
			content.append(System.getProperty("line.separator"));
		}
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(file));
		bw.write(content.toString());
		bw.close();
	}

	@Override
	public void saveRemovedZones(SlaveGroup slaveGroup) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving removed zones...");
		}
		
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(REMOVED_EXTENSION);
		
		File file = new File(path.toString());
		StringBuilder content = new StringBuilder();

		for (Zonename zonename : slaveGroup.getRemovedZonenames()) {
			content.append(zonename.getName());
			content.append(System.getProperty("line.separator"));
		}
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(file));
		bw.write(content.toString());
		bw.close();
		
	}

	@Override
	public void saveAddedCommand(SlaveGroup slaveGroup, String timestamp) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving added command...");
		}
		
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(ADDED_COMMAND_EXTENSION);
		path.append(timestamp);
		
		File file = new File(path.toString());
		StringBuilder content = new StringBuilder();

		InetAddress localhost = InetAddress.getLocalHost();
		if (localhost.isLoopbackAddress()) {
			logger.warn("Could not determine local IP address. Using loopback");
		}
		
		XfrdConfig xfrdConfig = XfrdConfigManager.getInstance();
		int port = xfrdConfig.getDnsServerConfig().getPort();

		String dnsPort = String.valueOf(port);
		for (Zonename zonename : slaveGroup.getAddedZonenames()) {
			if (slaveGroup.getRemovedZonenames().contains(zonename)) {
				continue;
			}
			content.append(zonename.getName());
			content.append(" '{ type slave; masters { ");
			content.append(localhost.getHostAddress());
			content.append(" port ");
			content.append(dnsPort);
			content.append("; }; file \"dnsshim");
			content.append(File.separatorChar);
			String hash = FileSystemUtil.hash(zonename.getName());
			content.append(hash);
			content.append(zonename.getName());
			content.append("\"; };'");
			content.append(System.getProperty("line.separator"));
		}
		if (content.length() > 0) {
			BufferedWriter bw = new BufferedWriter(new FileWriter(file));
			bw.write(content.toString());
			bw.close();
		}
	}

	@Override
	public void saveRemovedCommand(SlaveGroup slaveGroup, String timestamp) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Saving removed command...");
		}
		
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(REMOVED_COMMAND_EXTENSION);
		path.append(timestamp);
		
		File file = new File(path.toString());
		StringBuilder content = new StringBuilder();

		for (Zonename zonename : slaveGroup.getRemovedZonenames()) {
			if (slaveGroup.getAddedZonenames().contains(zonename)) {
				continue;
			}
			content.append(zonename.getName());
			content.append(System.getProperty("line.separator"));
		}
		
		if (content.length() > 0) {
			BufferedWriter bw = new BufferedWriter(new FileWriter(file));
			bw.write(content.toString());
			bw.close();
		}
	}

	@Override
	public void loadSyncErrors(SlaveGroup slaveGroup) throws IOException {
		if (slaveGroup == null) {
			return;
		}
		
		File groupPath = new File(BASE_DIR);
		File[] list; 
		
		list = groupPath.listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return (name.endsWith(ADDED_ERROR_EXTENSION));
			}
		});
		
		if (list.length > 0) {
			logger.info("No error files");
		}
		
		for (File file : list) {
			File tmp = new File(file.getAbsolutePath()+".tmp");
			file.renameTo(tmp);
			parseAddedZoneFile(tmp, slaveGroup);
			file.delete();
			tmp.delete();
		}
		
		list = groupPath.listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return (name.endsWith(REMOVED_ERROR_EXTENSION));
			}
		});
		
		if (list.length > 0) {
			logger.info("No error files");
		}
		
		for (File file : list) {
			File tmp = new File(file.getAbsolutePath()+".tmp");
			file.renameTo(tmp);
			parseRemovedZoneFile(tmp, slaveGroup);
			file.delete();
			tmp.delete();
		}
	}

	@Override
	public boolean deleteAddedZones(SlaveGroup slaveGroup) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(ADDED_EXTENSION);
		
		File file = new File(path.toString());
		return file.delete();
	}

	@Override
	public boolean deleteRemovedZones(SlaveGroup slaveGroup) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(slaveGroup.getName());
		path.append(REMOVED_EXTENSION);
		
		File file = new File(path.toString());
		return file.delete();
	}

	@Override
	public String getBaseDir() {
		return BASE_DIR;
	}
}