/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.scheduler;

import java.util.Calendar;
import java.util.Date;
import java.util.Random;
import java.util.TimeZone;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.log4j.Logger;

import br.registro.dnsshim.xfrd.domain.XfrdConfig;
import br.registro.dnsshim.xfrd.domain.logic.XfrdConfigManager;
import br.registro.dnsshim.xfrd.scheduler.domain.ZoneResign;
import br.registro.dnsshim.xfrd.service.PublicationService;
import br.registro.dnsshim.xfrd.service.PublicationServiceImpl;
import br.registro.dnsshim.xfrd.ui.protocol.PubZoneResponse;

public class ResignScheduler implements Runnable {
	private static final Logger logger = Logger.getLogger(ResignScheduler.class);
	private ExecutorService executorService = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
	private final CompletionService<PubZoneResponse> completionService = new ExecutorCompletionService<PubZoneResponse>(executorService);
	
	private final PriorityBlockingQueue<ZoneResign> queue;
	
	private int rescheduledCount; // amount of zones that has been rescheduled
	
	private int firstBoundary; 
	private int secondBoundary;
	
	private int rescheduleRange;
	
	private Random random = new Random();
	
	public ResignScheduler(PriorityBlockingQueue<ZoneResign> q) {
		this.queue = q;
		XfrdConfig xfrdConfig = XfrdConfigManager.getInstance();
			
		this.firstBoundary = xfrdConfig.getSchedulerHighLimit();
		this.secondBoundary = xfrdConfig.getSchedulerLowLimit();
		if (firstBoundary >= secondBoundary) {
			throw new IllegalArgumentException("Low priority resign boundary cannot execute before high priority boundary. Change xfrd config.");
		}
		rescheduleRange = secondBoundary - firstBoundary;
	}

	@Override
	public void run() {
		while (true) {
			try {
				if (queue.isEmpty()) {
					if (logger.isDebugEnabled()) {
						logger.debug("Resign Scheduler's queue empty.Waiting...");
					}
					synchronized (queue) {
						queue.wait();
					}
				} else {
					
					TimeZone timeZoneUTC = TimeZone.getTimeZone("UTC");
					TimeZone.setDefault(timeZoneUTC);
					Calendar calendar = Calendar.getInstance();
					
					Date now = calendar.getTime();
					
					calendar.add(Calendar.HOUR, firstBoundary);
					Date highPriorityBoudary = calendar.getTime();
					
					calendar.setTime(now);
					calendar.add(Calendar.HOUR, secondBoundary);
					Date lowPriorityBoundary = calendar.getTime();
					
					ZoneResign signature = queue.peek();
					
					if (signature.getScheduledDate() > 0) {
						if (logger.isDebugEnabled()) {
							logger.debug(signature.getZonename() + " resigning scheduled for "+ new Date(signature.getScheduledDate()));
						}
					} else {
						if (logger.isDebugEnabled()) {
							logger.debug(signature.getZonename() + " signature expiration on "+new Date(signature.getExpiration()));
						}
					}
					
					/* High priority condition...resigning now */
					if (highPriorityBoudary.getTime() >= signature.getExpiration()) {
						if (logger.isInfoEnabled()) {
							logger.info("Resigning " + signature.getZonename() + " now (Expiration on " + new Date(signature.getExpiration()) + ')');
						}
				
						signature = queue.take(); // removes from queue
						completionService.submit(new Resigner(signature.getZonename()));
						
					/* if signature has been rescheduled and signature or scheduled date has expired...resign now */	
					} else if (signature.getScheduledDate() > 0 && now.getTime() >= signature.getScheduledDate()) {
						if (logger.isInfoEnabled()) {
							logger.info("Starting "+ signature.getZonename() + "'s resigning now.(Expiration on " + new Date(signature.getExpiration()) + ')');
						}
						signature = queue.take(); // removes from queue
						completionService.submit(new Resigner(signature.getZonename()));
						
					/* if signature has not been rescheduled and signature hasn't expired, but will in the next secondBoundary hours...
					 * reschedule it */	
					} else if (signature.getScheduledDate() == 0 && lowPriorityBoundary.getTime() >= signature.getExpiration()) {
						rescheduledCount++;
						calendar = Calendar.getInstance();
						calendar.add(Calendar.HOUR, rescheduledCount % rescheduleRange);
						int randomMinute = random.nextInt(59);
						calendar.set(Calendar.MINUTE, randomMinute);
						signature = queue.take();
						signature.setScheduledDate(calendar.getTime().getTime());
						queue.put(signature);
						if (logger.isDebugEnabled()) {
							logger.debug(signature.getZonename() + "'s signature will expire in less than "+secondBoundary+" hours.Resigning scheduled for " + new Date(signature.getScheduledDate()) + ')');
						}
					} else {
						calendar = Calendar.getInstance();
						// next hour
						calendar.add(Calendar.HOUR, 1);
						calendar.set(Calendar.MINUTE, 0);
						calendar.set(Calendar.SECOND, 0);
						Date nextHour = calendar.getTime();
						
						long firstOccurence = signature.getExpiration() < signature.getScheduledDate() ? signature.getExpiration() : signature.getScheduledDate(); 
						long minor = firstOccurence < nextHour.getTime() ? firstOccurence : nextHour.getTime();
						
						if (minor < now.getTime()) {
							if (logger.isDebugEnabled()) {
								logger.debug("Waiting until next hour");
							}
							synchronized (queue) {
								queue.wait(nextHour.getTime() - now.getTime());
							}
						} else {	
							if (logger.isDebugEnabled()) {
								logger.debug("Nothing to resign now.Waiting until "+new Date(minor));
							}
							synchronized (queue) {
								queue.wait(minor - now.getTime());
							}
						}
					}
				}
			} catch (Exception e) {
				logger.error(e.getMessage(), e);
			}
		}
	}
}

class Resigner implements Callable<PubZoneResponse> {
	private String zonename;

	public Resigner(String zonename) {
		this.zonename = zonename;
	}

	@Override
	public PubZoneResponse call() throws Exception {
		PublicationService service = new PublicationServiceImpl();
		return service.full(zonename, 0);
	}
	
}
